import fs from "fs-extra";
import path from "node:path";
import xml from "xml-js";
import { fileURLToPath } from "node:url";
import StackUtils from "stack-utils";
import figures from "figures";

import colors from "./colors.js";
import formatSerializedError from "./format-serialized-error.js";
import improperUsageMessage from "./improper-usage-messages.js";
import beautifyStack from "./beautify-stack.js";
import codeExcerpt from "../code-excerpt.js";

const nodeInternals = StackUtils.nodeInternals();

const removeAnsiColoring = (text) =>
	text.replace(
		/[\u001b\u009b][[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-ORZcf-nqry=><]/g,
		""
	);

export default class JunitReporter {
	constructor(options) {
		this.projectDir = options.projectDir;
		this.fileName = options.fileName;
		this.tests = new Map();

		this.relativeFile = (file) => {
			if (file.startsWith("file://")) {
				file = fileURLToPath(file);
			}

			return path.relative(options.projectDir, file);
		};
	}

	startRun(plan) {
		plan.status.on("stateChange", (evt) => this.consumeStateChange(evt));
	}

	async endRun() {
		const testCases = [...this.tests.values()];
		const testSuiteStats = {
			failures: 0,
			skipped: 0,
			time: 0,
			todo: 0,
			tests: testCases.length,
			timestamp: new Date().toISOString(),
		};

		for (const testCase of testCases) {
			switch (testCase.status) {
				case "skipped": {
					testSuiteStats.skipped += 1;
					break;
				}
				case "todo": {
					testSuiteStats.todo += 1;
					break;
				}
				case "failed": {
					testSuiteStats.failures += 1;
					break;
				}
			}

			if (testCase.durationInMs !== -1) {
				testSuiteStats.time += testCase.durationInMs / 1000;
			}
		}

		const junitXmlResult = {
			_declaration: {
				_attributes: {
					version: "1.0",
					encoding: "utf-8",
				},
			},
			testsuites: {
				testsuite: {
					_attributes: {
						name: "ava",
						...testSuiteStats,
					},
					testcase: testCases.map((testCase) => {
						const result = {
							_attributes: {
								classname: path.basename(testCase.testFile).split(".")[0],
								file: testCase.testFile,
								name: testCase.title,
								time:
									testCase.durationInMs === -1
										? 0
										: testCase.durationInMs / 1000,
							},
						};

						switch (testCase.status) {
							case "skipped": {
								result.skipped = {
									_attributes: {
										message: "",
									},
								};
								break;
							}
							case "todo": {
								result.skipped = {
									_attributes: {
										message: "TODO:",
									},
								};
								break;
							}
							case "failed": {
								result.failure = {
									_text: removeAnsiColoring(testCase.errors.join("")),
								};
								break;
							}
						}

						const systemOut = removeAnsiColoring(testCase.logs.join(""));
						if (systemOut !== "") {
							result["system-out"] = {
								_text: systemOut,
							};
						}

						return result;
					}),
				},
			},
		};

		const directory = path.dirname(this.fileName);
		await fs.mkdirp(directory);

		await fs.writeFile(
			this.fileName,
			xml.js2xml(junitXmlResult, { compact: true }),
			"utf-8"
		);
	}

	consumeStateChange(evt) {
		// eslint-disable-line complexity
		// console.log(JSON.stringify(evt));
		switch (evt.type) {
			case "declared-test": {
				this.tests.set(`${evt.testFile}::${evt.title}`, {
					title: evt.title,
					testFile: path.relative(this.projectDir, evt.testFile),
					status: undefined,
					logs: [],
					durationInMs: -1,
				});
				break;
			}
			case "selected-test": {
				const test = this.tests.get(`${evt.testFile}::${evt.title}`);

				if (evt.skip) {
					test.status = "skipped";
				} else if (evt.todo) {
					test.status = "todo";
				}

				break;
			}
			case "test-failed": {
				const test = this.tests.get(`${evt.testFile}::${evt.title}`);

				test.logs = evt.logs;
				test.durationInMs = evt.duration;
				test.status = "failed";
				test.errors = [this.getPrintableError(evt)];

				break;
			}
			case "test-passed": {
				const test = this.tests.get(`${evt.testFile}::${evt.title}`);

				test.logs = evt.logs;
				test.durationInMs = evt.duration;
				test.status = "passed";

				break;
			}
		}
	}

	getPrintableError(event) {
		const errorLines = [];

		if (
			event.err.name === "TSError" &&
			event.err.object &&
			event.err.object.diagnosticText
		) {
			errorLines.push(colors.errorStack(event.err.object.diagnosticText));
			errorLines.push("");
			return errorLines.join("\n");
		}

		if (event.err.source) {
			errorLines.push(
				colors.errorSource(
					`${this.relativeFile(event.err.source.file)}:${event.err.source.line}`
				)
			);
			const excerpt = codeExcerpt(event.err.source, {
				maxWidth: 82,
			});
			if (excerpt) {
				errorLines.push("");
				errorLines.push(excerpt);
				errorLines.push("");
			}
		}

		if (event.err.avaAssertionError) {
			const result = formatSerializedError(event.err);
			if (result.printMessage) {
				errorLines.push(event.err.message);
				errorLines.push("");
			}

			if (result.formatted) {
				errorLines.push(result.formatted);
				errorLines.push("");
			}

			const message = improperUsageMessage(event.err);
			if (message) {
				errorLines.push(message);
				errorLines.push("");
			}
		} else if (event.err.nonErrorObject) {
			errorLines.push(event.err.formatted);
			errorLines.push("");
		} else {
			errorLines.push(event.err.summary);
			errorLines.push("");
		}

		const formatted = this.formatErrorStack(event.err);
		if (formatted.length > 0) {
			errorLines.push(formatted.join("\n"));
			errorLines.push("");
		}

		return errorLines.join("\n");
	}

	formatErrorStack(error) {
		if (!error.stack) {
			return [];
		}

		if (error.shouldBeautifyStack) {
			return beautifyStack(error.stack).map((line) => {
				if (nodeInternals.some((internal) => internal.test(line))) {
					return colors.errorStackInternal(`${figures.pointerSmall} ${line}`);
				}

				return colors.errorStack(`${figures.pointerSmall} ${line}`);
			});
		}

		return [error.stack];
	}
}

// {"type":"declared-test","title":"b2","knownFailing":false,"todo":false,"testFile":"/home/martynas/projects/github/test-cases/src/ava/b.int.ts"}
// {"type":"selected-test","title":"b2","knownFailing":false,"skip":true,"todo":false,"testFile":"/home/martynas/projects/github/test-cases/src/ava/b.int.ts"}
//
